----------------------------------------------------------------------
----                                                              ----
---- ATARI DMA compatible IP Core					              ----
----                                                              ----
---- This file is part of the SUSKA ATARI clone project.          ----
---- http://www.experiment-s.de                                   ----
----                                                              ----
---- Description:                                                 ----
---- ATARI ST and STE compatible DMA controller IP core.          ----
----                                                              ----
---- This files is moddeling the DMA relevant registers.          ----
----                                                              ----
----                                                              ----
---- To Do:                                                       ----
---- -                                                            ----
----                                                              ----
---- Author(s):                                                   ----
---- - Wolfgang Foerster, wf@experiment-s.de; wf@inventronik.de   ----
----                                                              ----
----------------------------------------------------------------------
----                                                              ----
---- Copyright (C) 2006 - 2011 Wolfgang Foerster                  ----
----                                                              ----
---- This source file may be used and distributed without         ----
---- restriction provided that this copyright statement is not    ----
---- removed from the file and that any derivative work contains  ----
---- the original copyright notice and the associated disclaimer. ----
----                                                              ----
---- This source file is free software; you can redistribute it   ----
---- and/or modify it under the terms of the GNU Lesser General   ----
---- Public License as published by the Free Software Foundation; ----
---- either version 2.1 of the License, or (at your option) any   ----
---- later version.                                               ----
----                                                              ----
---- This source is distributed in the hope that it will be       ----
---- useful, but WITHOUT ANY WARRANTY; without even the implied   ----
---- warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ----
---- PURPOSE. See the GNU Lesser General Public License for more  ----
---- details.                                                     ----
----                                                              ----
---- You should have received a copy of the GNU Lesser General    ----
---- Public License along with this source; if not, download it   ----
---- from http://www.gnu.org/licenses/lgpl.html                   ----
----                                                              ----
----------------------------------------------------------------------
-- 
-- Revision History
-- 
-- Revision 2K6A  2006/06/03 WF
--   Initial Release.
-- Revision 2K6B  2006/11/06 WF
--   Modified Source to compile with the Xilinx ISE.
-- Revision 2K8A  2008/07/14 WF
--   Introduced the access level register CD_2_DATA.
-- Revision 2K8B  2008/12/24 WF
--   Introduced SCSICSn and SDCSn.
--   Introduced DMA_SRC_SEL as a bit vector.
--   DMA_STATUS_REG has now synchronous reset to meet preset requirement.
-- Revision 2K9B  2009/12/24 WF
--   Changed timing of SECT_CNT_ZEROn.
--   Fixed a bug in the sector counter.
--   Fixed DMA_EN logic and replaced DMA_RDn, DMA_WRn by DMA_EN.
--   Introduced CTRL_SRC_SEL.
--   Fixed bus access timing.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity WF25913IP_REGISTERS is
	port(
		CLK				: in bit;
		RESETn			: in bit;
		FCSn			: in bit;
		RWn				: in bit;
		A1  			: in bit;

		FIFO_ERROR		: in bit;
		ACSI_DATA_REQ	: in bit;
		SECTOR_CNT_EN	: in bit;
		
	    DATA_IN			: in std_logic_vector (8 downto 0);
	    DATA_OUT		: out std_logic_vector (15 downto 0);
		DATA_EN			: out bit;
		CD_IN			: in std_logic_vector (7 downto 0);
		CD_OUT			: out std_logic_vector (7 downto 0);
		CD_EN			: out bit;		

		CTRL_SRC_SEL	: out bit_vector(1 downto 0);
		DMA_SRC_SEL		: out bit_vector(1 downto 0);
		DMA_EN          : out bit;

		DMA_RWn			: out bit;
        HDCSn			: out bit;
        SCSICSn         : out bit;
        SDCSn           : out bit;
        FDCSn			: out bit;
		CA				: out bit_vector(1 downto 0);
		CTRL_ACC		: out bit
	);
end WF25913IP_REGISTERS;

architecture BEHAVIOR of WF25913IP_REGISTERS is
signal DMA_STATUS_REG	: bit_vector(2 downto 0);
signal DMA_MODE_REG		: bit_vector(8 downto 0);
signal SECTOR_CNT_REG	: std_logic_vector(7 downto 0);
signal SECT_CNT_ZEROn   : bit;
signal CTRL_ACC_EN		: bit;
signal HDCSn_I			: bit;
signal FDCSn_I			: bit;
signal SCSICSn_I		: bit;
signal SDCSn_I			: bit;
signal CTRL_MASK		: std_logic_vector(1 downto 0);
begin
	DMA_MODE: process(RESETn, CLK)
	-- The DMA mode register is write only. At the same 
	-- adress the DMA status register is read only.
	begin
		if RESETn = '0' then
			DMA_MODE_REG <= (others => '0');
		elsif CLK = '1' and CLK' event then
			if FCSn = '0' and A1 = '1' and RWn = '0' then
                DMA_MODE_REG <= To_BitVector(DATA_IN); -- Write to register.
			end if;
		end if;
	end process DMA_MODE;

	-- Wiring of the DMA mode register ($FF8606|word, W):
	--  DMA mode/status                 BIT 8 7 6 5 4 3 2 1 0
	--  0 - read FDC/HDC, 1 - write --------' | | | | | | | |
	--  0 - DMA with HDC/SCSI, 1 - with FDC/SD' | | | | | | |
	--  0 - DMA on, 1 - no DMA -----------------' | | | | | |
	--  ACSI/SCSI/SD/Floppy see below ------------' | | | | |
	--  0 - FDC reg, 1 - sector count reg ----------' | | | |
	--  ACSI/SCSI/SD/Floppy see below ----------------' | | |
	--  0 - pin CA1 low, 1 - pin CA1 high --------------' | |
	--  0 - pin CA0 low, 1 - pin CA0 high ----------------' |
    --  0 - DMA with HDC/FDC, 1 - DMA with SCSI/SD-Card ----'

    -- Selection for the DMA_SRC_SEL(1) Bit 7 and DMA_SRC_SEL(0) Bit 5 of the DMA mode register:
    -- 00   : DMA with ACSI:
    -- 10   : DMA with FLOPPY:
    -- 01   : DMA with SCSI:
    -- 11   : DMA with SD card.

	DMA_RWn <= not DMA_MODE_REG(8);
    DMA_SRC_SEL(1) <= DMA_MODE_REG(7);
    DMA_SRC_SEL(0) <= DMA_MODE_REG(5);
	DMA_EN <= not DMA_MODE_REG(6);
    CTRL_ACC_EN <= not DMA_MODE_REG(4);
	CA <= DMA_MODE_REG(2 downto 1);

	DMA_STATUS: process
	-- The DMA status register is read only. At the same 
	-- adress the DMA mode register is write only.
	begin
		wait until CLK = '1' and CLK' event;
		if RESETn = '0' then
            DMA_STATUS_REG <= "011"; -- Register is low active.
		-- Clear the status register by access (read or write)
		-- to the sector count register:
		elsif FCSn = '0' and A1 = '0' and CTRL_ACC_EN = '0' then
            DMA_STATUS_REG <= "111"; -- Clear.
		elsif FIFO_ERROR = '1' then
			DMA_STATUS_REG(0) <= '0'; -- Store the event.
		else
            DMA_STATUS_REG(2) <= ACSI_DATA_REQ; -- Update.
			DMA_STATUS_REG(1) <= SECT_CNT_ZEROn; -- Update.
		end if;
	end process DMA_STATUS;
	
	SECTOR_CNT: process(RESETn, CLK, SECTOR_CNT_REG)
	begin
		if RESETn = '0' then
			SECTOR_CNT_REG <= x"00";
		elsif CLK = '1' and CLK' event then
			if FCSn = '0' and A1 = '0' and RWn = '0' and CTRL_ACC_EN = '0' then
				SECTOR_CNT_REG <= DATA_IN(7 downto 0); -- Write to register.
            elsif SECTOR_CNT_EN = '1' and SECTOR_CNT_REG > x"00" then
				SECTOR_CNT_REG <= SECTOR_CNT_REG - '1'; -- Count down.
			end if;
		end if;
	end process SECTOR_CNT;

    -- Bring the SECTOR_CNT_ZEROn information as early as possible drung DMA write to target.
    SECT_CNT_ZEROn <= '0' when SECTOR_CNT_REG = x"01" and SECTOR_CNT_EN = '1' and DMA_MODE_REG(8) = '0' else
                      '0' when SECTOR_CNT_REG = x"00" else '1';

	-- Read from register:
	-- In read operation unused pins are read back as '0's:
	DATA_EN <= '1' when FCSn = '0' and RWn = '1' else '0';
	DATA_OUT <=	"0000000000000" & To_StdLogicVector(DMA_STATUS_REG) when FCSn = '0' and A1 = '1' and RWn = '1' else
				x"00" & SECTOR_CNT_REG when FCSn = '0' and A1 = '0' and RWn = '1' and CTRL_ACC_EN = '0' else
				-- Controller access stuff:
                x"00" & CD_IN when FCSn = '0' and A1 = '0' and RWn = '1' and CTRL_ACC_EN = '1' else (others => '0');

	CTRL_ACC <= '1' when CTRL_ACC_EN = '1' and FCSn = '0' and A1 = '0' else '0'; -- Controller access.

    CD_EN <= '1' when CTRL_ACC_EN = '1' and FCSn = '0' and A1 = '0' and RWn = '0' else '0';
    CD_OUT <= DATA_IN(7 downto 0);

	CTRL_ACC_TIMING: process (CLK, RESETn)
	-- This process provides the original timing to the WD1772 asserting several control
	-- signals on the ACSI bus. The timing is triggered by an access of the controller
	-- access register and stops after the mask counter ends its count cycle.
	-- This timing must be correlated with the RDY_OUTn timing in the control file!
	variable LOCK : boolean;
	begin
		if RESETn = '0' then
            CTRL_MASK <= "00";
			LOCK := false;
        elsif CLK = '1' and CLK' event then
			if CTRL_ACC_EN = '1' and FCSn = '0' and A1 = '0' and LOCK = false then
                CTRL_MASK <= "11"; -- Start condition immediately after the bus access.
				LOCK := true;
            elsif CTRL_MASK > "00" then
				CTRL_MASK <= CTRL_MASK - 1; -- Active.
            elsif CTRL_MASK = "00" and FCSn = '1' then
				LOCK := false;	-- Reset condition.
			end if;
		end if;
	end process CTRL_ACC_TIMING;

    -- Controller access drive selection:
    -- This configuration was chosen due to compatibility with old software.
    -- Use the same configuration as for the DMA decoding to mix CTRL_SRC_SEL
    -- easily up with DMA_SRC_SEL in the DMA top level file.
    with DMA_MODE_REG(3) & DMA_MODE_REG(0) select
        CTRL_SRC_SEL <= "00" when "10", -- ACSI. 
                        "01" when "11", -- SCSI.
                        "10" when "00", -- Floppy.
                        "11" when "01"; -- SD card.

    FDCSn_I <= '0' when DMA_MODE_REG(3) = '0' and DMA_MODE_REG(0) = '0' else '1';
    HDCSn_I <= '0' when DMA_MODE_REG(3) = '1' and DMA_MODE_REG(0) = '0' else '1';
    SDCSn_I <= '0' when DMA_MODE_REG(3) = '0' and DMA_MODE_REG(0) = '1' else '1';
    SCSICSn_I <= '0' when DMA_MODE_REG(3) = '1' and DMA_MODE_REG(0) = '1' else '1';

    with CTRL_MASK select  HDCSn <= HDCSn_I when "11" | "10" | "01", '1' when others; -- Phases 3, 2, 1.
    with CTRL_MASK select  SCSICSn <= SCSICSn_I when "11" | "10" | "01", '1' when others; -- Phases 3, 2, 1.
    with CTRL_MASK select  SDCSn <= SDCSn_I when "11" | "10" | "01", '1' when others; -- Phases 3, 2, 1.
    with CTRL_MASK select  FDCSn <= FDCSn_I when "11" | "10" | "01", '1' when others; -- Phases 3, 2, 1.
end architecture BEHAVIOR;
